#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

__docformat__ = 'restructuredtext'


# import standard library

# import interfaces
from pyams_pagelet.interfaces import PageletCreatedEvent, IPagelet
from pyams_portal.interfaces import IPortalContext, IPortalPage, IPortalTemplateConfiguration
from pyams_skin.layer import IPyAMSLayer
from pyams_utils.interfaces import MANAGE_PERMISSION
from pyams_workflow.interfaces import IWorkflowVersions, IWorkflowState
from pyams_zmi.interfaces.menu import ISiteManagementMenu, IPropertiesMenu
from pyams_zmi.layer import IAdminLayer

# import packages
from pyramid.exceptions import NotFound
from pyramid.view import view_config
from pyams_form.form import AJAXEditForm
from pyams_pagelet.pagelet import pagelet_config
from pyams_portal.workflow import PUBLISHED, ARCHIVED
from pyams_portal.zmi.template.config import PortalTemplateConfigView
from pyams_skin.viewlet.menu import MenuItem
from pyams_utils.url import absolute_url
from pyams_viewlet.viewlet import viewlet_config
from pyams_zmi.form import AdminDialogEditForm
from z3c.form import field

from pyams_portal import _



@viewlet_config(name='template-properties.menu', context=IPortalContext, layer=IAdminLayer,
                manager=IPropertiesMenu, permission=MANAGE_PERMISSION, weight=5)
class PortalContextTemplatePropertiesMenu(MenuItem):
    """Portal context template properties menu"""

    label = _("Presentation template...")
    icon_class = 'fa-columns'

    url = 'template-properties.html'
    modal_target = True


@pagelet_config(name='template-properties.html', context=IPortalContext, layer=IPyAMSLayer,
                permission=MANAGE_PERMISSION)
class PortalContextTemplatePropertiesEditForm(AdminDialogEditForm):
    """Portal context template properties edit form"""

    @property
    def title(self):
        return self.context.title

    legend = _("Edit template configuration")

    @property
    def fields(self):
        fields = field.Fields(IPortalPage).select('inherit_parent', 'use_local_template', 'shared_template')
        if not self.getContent().can_inherit:
            fields = fields.omit('inherit_parent')
        return fields

    ajax_handler = 'template-properties.json'
    edit_permission = MANAGE_PERMISSION

    def getContent(self):
        return IPortalPage(self.context)


@view_config(name='template-properties.json', context=IPortalContext, request_type=IPyAMSLayer,
             permission=MANAGE_PERMISSION, renderer='json', xhr=True)
class PortalContextTemplatePropertiesAJAXEditForm(AJAXEditForm, PortalContextTemplatePropertiesEditForm):
    """Portal context template properties edit form, JSON renderer"""


@viewlet_config(name='template-config.menu', context=IPortalContext, layer=IAdminLayer,
                manager=ISiteManagementMenu, permission=MANAGE_PERMISSION, weight=20)
class PortalContextTemplateConfigMenu(MenuItem):
    """Portal context template configuration menu"""

    label = _("Template properties")
    icon_class = 'fa-columns'

    url = '#template-config.html'

    def __new__(cls, context, request, view, manager=None):
        page = IPortalPage(context)
        if page.template is None:
            return None
        return MenuItem.__new__(cls)

    def get_url(self):
        page = IPortalPage(self.context)
        if page.use_local_template:
            template = IWorkflowVersions(page.template).get_last_versions()[0]
            return absolute_url(template, self.request, 'admin.html#properties.html')
        else:
            return super(PortalContextTemplateConfigMenu, self).get_url()


@pagelet_config(name='template-config.html', context=IPortalContext, layer=IPyAMSLayer,
                permission=MANAGE_PERMISSION)
class PortalContextTemplateConfigView(PortalTemplateConfigView):
    """Portal context template configuration view"""

    title = _("Local portal template configuration")

    def get_context(self):
        template = IPortalPage(self.context).template
        return IWorkflowVersions(template).get_last_versions()[0]

    @property
    def can_change(self):
        if not IPortalPage(self.context).use_local_template:
            return False
        return self.request.has_permission(MANAGE_PERMISSION) and \
               IWorkflowState(self.get_context()).state not in (PUBLISHED, ARCHIVED)


@view_config(name='portlet-properties.html', context=IPortalContext, request_type=IPyAMSLayer,
             permission=MANAGE_PERMISSION)
class PortalContextTemplatePortletEditForm(AdminDialogEditForm):
    """Portal context template portlet edit form"""

    dialog_class = 'modal-large'

    def __call__(self):
        request = self.request
        request.registry.notify(PageletCreatedEvent(self))
        slot_name = request.params.get('form.widgets.slot_name')
        position = int(request.params.get('form.widgets.position'))
        config = IPortalTemplateConfiguration(self.context)
        portlet_config = config.get_portlet_configuration(slot_name, position)
        if portlet_config is None:
            raise NotFound()
        editor = self.request.registry.queryMultiAdapter((portlet_config, request),
                                                         IPagelet, name='properties.html')
        if editor is None:
            raise NotFound()
        editor.ajax_handler = 'portlet-properties.json'
        editor.update()
        return editor()
